/*-
 * Copyright Sergey Kosyakov ks@itp.ac.ru 1999
 * Copyright (C) @BABOLO  2002 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright Sergey Kosyakov ks@itp.ac.ru 1999\n\
@(#)Copyright (C) @BABOLO  2002 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: utils.c,v 1.7 2010/10/20 05:29:48 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/file.h>
#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include "tund.h"

#define SEQ_INISIZE	16

extern int debug;
extern int control_port;
extern Seq *stun;

/*****************************************************************************/
/* Sequence support                                                          */
/*****************************************************************************/

Seq *
seq_new(tund_conf *cnf) {
    ifBLIN_FLAG = cnf->flags;
    Seq *ret;

    if  ((ret = calloc(1, sizeof(Seq)))) {
        ret->maxlen = SEQ_INISIZE;
        ret->length = 0;
        if  (!(ret->buf = calloc(SEQ_INISIZE, sizeof(void *)))) {
            ifBLIN_QX0("no mem for Seq->buf");
            free(ret);
            ret = NULL;
            errno = ENOMEM;
        }
    } else {
        ifBLIN_QX0("no mem for Seq");
        errno = ENOMEM;
    }
    return(ret);
}

int
seq_append(tund_conf *cnf, Seq *s, void *item) {
    void **buf;
    int len;

    s->length++;
    if  (s->length >= s->maxlen) {
        len = s->maxlen * 2;
        buf = calloc(len, sizeof(void *));
        /* XXXX check buf == NULL */
        bcopy(s->buf, buf, s->maxlen * sizeof(void *));
        free(s->buf);
        s->buf = buf;
        s->maxlen = len;
    }
    s->buf[s->length - 1] = item;
    return(s->length - 1);
}

/*****************************************************************************/
/* Log and errors                                                            */
/*****************************************************************************/

#define LOG_BUFS	1024
static char log_buf[LOG_BUFS];
static int log_init = 0;

void
Log(char *fmt, ...) {
    va_list ap;
    int ret;

    if  (!log_init) {
        openlog("tund", LOG_CONS | LOG_NDELAY, LOG_DAEMON);
        log_init = 1;
    }
    bzero(log_buf, LOG_BUFS);
    va_start(ap, fmt);
    ret = vsnprintf(log_buf, LOG_BUFS, fmt, ap);
    va_end(ap);
    if(debug) printf("%s\n", log_buf); else syslog(LOG_WARNING, log_buf);
}

void
Error(char *fmt, ...) {
    va_list ap;
    int ret;

    if  (!log_init) {
        openlog("tund", LOG_CONS | LOG_NDELAY, LOG_DAEMON);
        log_init = 1;
    }
    bzero(log_buf, LOG_BUFS);
    va_start(ap, fmt);
    ret = vsnprintf(log_buf, LOG_BUFS, fmt, ap);
    va_end(ap);
    if  (debug) printf("%s\n", log_buf); else syslog(LOG_ERR, log_buf);
    exit(1);
}

/*****************************************************************************/
/* Parsing config file                                                       */
/*****************************************************************************/

static int
parse_host_name(tund_conf *cnf, char *name, struct in_addr *addr) { 
    int n1, n2, n3, n4;
    char buf[32];
    char dummy;
    int ret;  

    ret = sscanf(name, "%d.%d.%d.%d%c", &n1, &n2, &n3, &n4, &dummy);
    if  (ret != 4) return(0);

    if  ((n1 < 0) || (n1 > 255)) return(0);
    if  ((n2 < 0) || (n2 > 255)) return(0);
    if  ((n3 < 0) || (n3 > 255)) return(0);
    if  ((n4 < 0) || (n4 > 255)) return(0);
    snprintf(buf, 32, "%d.%d.%d.%d", n1, n2, n3, n4);
    addr->s_addr = inet_addr(buf);
    return(1);
}

static int
parse_config_options(tund_conf *cnf, Tunnel *tun, char *str) {
    struct in_addr addr;
    struct hostent *hs;

    while (*str == ' ' || *str == '\n' || *str == '\t') str++;
    if  (!(*str)) return(0);
    if  (strncmp(str, "rule=", 5) == 0) {
        tun->tag = atoi(str + 5);
        return(0);
    }
    if  (strncmp(str, "divert=", 7) == 0) {
        tun->dp = atoi(str + 7);
        return(0);
    }
    if  (strncmp(str, "host=", 5) == 0) {
        bzero(&(tun->to), sizeof(struct sockaddr_in));
        tun->to.sin_port = htons((unsigned short)control_port);
        tun->to.sin_family = AF_INET;
        if  (parse_host_name(cnf, str + 5, &addr)) {
            bcopy(&addr, &(tun->to.sin_addr), sizeof(struct in_addr));
        } else {
            hs = gethostbyname(str + 5);
            if  (!hs) {
                printf("Can not find host %s\n", str + 5);
                return(-1);
            }
            bcopy(hs->h_addr_list[0], &addr, sizeof(addr));
            bcopy(&addr, &(tun->to.sin_addr), sizeof(struct in_addr));
        }
        tun->remoteName = strdup(str + 5);
        return(0);
    }
    if  (strncmp(str, "encrypt=", 8) == 0) {
        tun->flags |= TUNF_ENC;
        if  (strcmp(str + 8, "blowfish") == 0) {
            tun->flags |= TUNF_CIPHER_BF;
        } else if(strcmp(str + 8, "idea") == 0) {
            tun->flags |= TUNF_CIPHER_IDEA;
        } else if(strcmp(str + 8, "rc5") == 0) {
            tun->flags |= TUNF_CIPHER_RC5;
        } else {
            Error("Unknown cipher [%s] for tunnel [%s]", str + 8, tun->label);
        }
        return(0);
    }
    return(1);
}

int
parse_config(tund_conf *cnf) {
    char str[1024];
    Tunnel *tunnel;
    int line = 0;
    char *tmp;
    char *p;
    FILE *f;

    f = fopen(cnf->conf_nm, "r");
    if  (!f) {
        printf("Can not open config file %s\n", cnf->conf_nm);
        return(-1);
    }
    while (fgets(str, 1023, f)) {
        p = str;
        line++;
        while(*p == ' ') p++;                                                         /* skip blanks */
        if  (*p == '#' || *p == '\n' || *p == '\r' || *p == 0) continue;
        if  (strncmp(p, "tunnel", 6) == 0) {
            p += 6;
            while (*p == ' ') p++;
            tunnel = calloc(1, sizeof(Tunnel));
            tunnel->tun_desc = strdup(p);                                             /* Tunnel line */
            tmp = strsep(&p, " \t");                                                        /* label */
            if  (!tmp) Error("Error in config file %s at line %d\n", cnf->conf_nm, line);
            tunnel->label = strdup(tmp);
            tunnel->tunAddr = atoi(tmp);
            while ((tmp = strsep(&p, " \t\n"))) {                                         /* options */
                if  (parse_config_options(cnf, tunnel, tmp)) {
                    Error("Illegal option [%s]", tmp);
            }   }
            tunnel->l_addr.sin_port = tunnel->tag;
            register_tunnel(cnf, tunnel->tunAddr, seq_append(cnf, stun, tunnel));
            if  (debug)
                Log( "Tunnel [%s]: tag=%d dp=%d host=[%s] flags=0x%x a=%05hu"
                   , tunnel->label
                   , tunnel->tag
                   , tunnel->dp
                   , tunnel->remoteName
                   , tunnel->flags
                   , tunnel->tunAddr
                   );
    }   } 
    fclose(f);
    if  (control_port && stun->length > 0) return(0);
    errno = EINVAL;
    return(-1);
}

/*****************************************************************************/
/* RNG                                                                       */
/*****************************************************************************/

static int32_t *rng_F;
static int32_t rng_S;
static int32_t rng_RL = 9689;
static int32_t rng_D = 471;
static u_int32_t *rng_R;
static u_int32_t rng_msk = 0x7FFFFFFF;

int
get_rand(tund_conf *cnf) {
    int tmp;
    int n;

    tmp = rng_R[*rng_F] ^ rng_R[rng_S];
    tmp &= rng_msk;
    rng_R[*rng_F] = tmp;
    (*rng_F)++;
    rng_S++;
    if  (*rng_F != rng_RL) {
        if  (rng_S == rng_RL) rng_S = 0;
    } else {
        *rng_F = 0;
    }
    if  (debug) {
        n = *rng_F - rng_D;
        if  (n < 0) n = n + rng_RL;
        if  (n != rng_S) {
            printf("n=%d, rng_S=%d\n", n, rng_S);
            exit(1);
    }   }
    /* msync(rng_F, rng_RL * 4 + 4, MS_SYNC); */
    return(tmp);
}

int
init_rng(tund_conf *cnf) {
    int fd, ex = EX_OK;
    struct stat sb;
    int ret;

    fd = open(cnf->seed_nm, O_RDWR);
    if  (fd < 0) {
        Error("Can not open file %s", cnf->seed_nm);
        exit(1);
    }
    ret = fstat(fd, &sb);
    if  (ret < 0) {
        perror("Can not get stat of seed file");
        exit(1);
    }
    if  (sb.st_size != rng_RL * 4 + 4) {
        printf("Illegal size %d of seed file\n", (int)sb.st_size);
        exit(1);
    } 
    rng_F = mmap(NULL, rng_RL * 4 + 4, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if  (rng_F == MAP_FAILED) {
        perror("Can not mmap seed file");
        exit(1);
    }
    if  (*rng_F >= rng_RL || *rng_F < 0) {
        printf("Corrupted RNG seed file\n");
        exit(1);
    }
    rng_R = (u_int32_t *)rng_F + 1;
    rng_S = *rng_F - rng_D;
    if  (rng_S < 0) rng_S = rng_RL + rng_S;
    if  (debug) printf("F=%d, S=%d\n", *rng_F, rng_S);
    /* for (ret = 0; ret < 200000; ret++) printf("%d\n", get_rand()); */
    return(ex);
}
